#ifndef _FIFO_BUFFER_H
#define _FIFO_BUFFER_H

/**
 * @file fifo-buffer.h
 * 
 * Exemple de tampon circulaire de type FIFO.
 */

/**
 * Structure décrivant l'état du FIFO.
 */
typedef struct {
	size_t size; /**< Taille totale du tampon. */
	size_t length; /**< Taille utilisée. */
	size_t rIndex; /**< Indice de la prochaine lecture. */
	size_t wIndex; /**< Indice de la prochaine écriture. */
	uint8_t status; /**< Octet d'état utilisable par l'application. */
	uint8_t *data; /**< Adresse du tampon (doit être alloué statiquement). */
} FifoState;

/**
 * Declare une variable de type FifoState et son tampon statique.
 *
 * Le tampon est vide quand rIndex == wIndex.
 *
 * bufferSize est la taille totale du tampon.
 */
#define FIFO_BUFFER(variableName, bufferSize) \
	static uint8_t variableName ## Data[bufferSize]; \
	FifoState variableName = { \
		.size = bufferSize, \
		.length = 0, \
		.rIndex = 0, \
		.wIndex = 0, \
		.status = 0, \
		.data = variableName ## Data, \
	};

/**
 * Renvoie le nombre d'octets utilisés dans le tampon.
 *
 * @param fifo Adresse du tampon.
 * @return Nombre d'octets utilisés.
 */
INLINE size_t fifoLength(FifoState *fifo) {
	return fifo->length;
}

/**
 * Vide le tampon.
 *
 * @param fifo Adresse du tampon.
 */
void fifoClear(FifoState *fifo);

/**
 * Ajoute un bloc d'octets au tampon.
 *
 * @param fifo Adresse du tampon.
 * @param data Adresse du bloc d'octets à écrire dans le tampon.
 * @param count Taille du bloc d'octets à écrire dans le tampon.
 * @return true si l'écriture a réussi, false si le tampon est inchangé.
 */
bool fifoWrite(FifoState *fifo, const void *data, size_t count);

/**
 * Extrait un bloc d'octets du tampon.
 *
 * @param fifo Adresse du tampon.
 * @param data Adresse de destination.
 * @param count Taille du bloc d'octets à extraire du tampon.
 * @return true si l'extraction a réussi, false si le tampon est inchangé.
 */
bool fifoRead(FifoState *fifo, void *data, size_t count);

/**
 * Détermine si le tampon est plein.
 *
 * @param fifo Adresse du tampon.
 * @return true si le tampon est plein.
 */
bool fifoIsFull(FifoState *fifo);

/**
 * Détermine le nombre d'octets utilisés dans le tampon.
 * Alias de fifoLength().
 *
 * @param fifo Adresse du tampon.
 * @return Nombre d'octets utilisés.
 */
INLINE size_t fifoBytesUsed(FifoState *fifo) {
	return fifoLength(fifo);
}

/**
 * Détermine le nombre d'octets disponibles dans le tampon.
 *
 * @param fifo Adresse du tampon.
 * @return Nombre d'octets disponibles.
 */
size_t fifoBytesFree(FifoState *fifo);

/**
 * Détermine si le tampon est vide.
 *
 * @param fifo Adresse du tampon.
 * @return true si le tampon est vide.
 */
INLINE bool fifoIsEmpty(FifoState *fifo) {
	return fifoLength(fifo) == 0;
}

#endif // _FIFO_BUFFER_H
