#include "project-defs.h"
#include "fifo-buffer.h"
#include "uart-api.h"

// On définit des tampons de 16 octets - choix purement arbitraire dans
// cet exemple. On pourrait également très bien avoir des tampons de
// taille différente en émission et en réception, c'est à déterminer
// en fonction des besoins de l'aplication.
#define UART_BUFFER_SIZE 16

FIFO_BUFFER(USART1_receiveBuffer, UART_BUFFER_SIZE)
FIFO_BUFFER(USART1_transmitBuffer, UART_BUFFER_SIZE)

typedef enum {
	STATUS_IDLE = 0,
	STATUS_SENDING,
} TransmitStatus;

__attribute__((interrupt("WCH-Interrupt-fast")))
void USART1_IRQHandler() {
	uint8_t c;

	if (USART_GetITStatus(USART1, USART_IT_RXNE) != RESET) {
		// On vient de recevoir un caractère, on le met dans le tampon
		// de réception.
		c = USART_ReceiveData(USART1);
		fifoWrite(&USART1_receiveBuffer, &c, 1);
	}

	if (USART_GetITStatus(USART1, USART_IT_TXE) != RESET) {
		// L'UART est prêt à envoyer un autre caractère.
		if (fifoRead(&USART1_transmitBuffer, &c, 1)) {
			// Tant qu'il reste des caractères à traiter dans le tampon
			// d'émission, on les lit et on les envoie.
			USART_SendData(USART1, c);
		} else {
			// Le flag d'interruption USART_IT_TXE ne peut être remis à zéro
			// que par l'envoi d'un caractère. Or, nous arrivons ici parce que
			// nous avons déjà tout envoyé. Nous devons donc supprimer les
			// interruptions USART_IT_TXE en attendant d'avoir de nouveaux
			// caractères à envoyer - voir uartSendBlock().
			USART_ITConfig(USART1, USART_IT_TXE, DISABLE);
			USART1_transmitBuffer.status = STATUS_IDLE;
		}
	}
}

void uartInitialise(uint32_t baudRate) {
	GPIO_InitTypeDef gpioInit = { 0 };

	RCC_APB2PeriphClockCmd(RCC_APB2Periph_GPIOD, ENABLE);

	// RX est sur PD6
	gpioInit.GPIO_Pin = GPIO_Pin_6;
	gpioInit.GPIO_Mode = GPIO_Mode_IN_FLOATING;
	GPIO_Init(GPIOD, &gpioInit);

	// TX est sur PD5 et est une alternate function
	gpioInit.GPIO_Pin = GPIO_Pin_5;
	gpioInit.GPIO_Mode = GPIO_Mode_AF_PP;
	gpioInit.GPIO_Speed = GPIO_Speed_50MHz;
	GPIO_Init(GPIOD, &gpioInit);

	USART_InitTypeDef usartInit = { 0 };

	RCC_APB2PeriphClockCmd(RCC_APB2Periph_USART1, ENABLE);

	usartInit.USART_BaudRate = baudRate;
	usartInit.USART_WordLength = USART_WordLength_8b;
	usartInit.USART_StopBits = USART_StopBits_1;
	usartInit.USART_Parity = USART_Parity_No;
	usartInit.USART_HardwareFlowControl = USART_HardwareFlowControl_None;
	usartInit.USART_Mode = USART_Mode_Tx | USART_Mode_Rx;
	USART_Init(USART1, &usartInit);

	USART_ITConfig(USART1, USART_IT_RXNE, ENABLE);

	NVIC_InitTypeDef  nvicInit = { 0 };

	nvicInit.NVIC_IRQChannel = USART1_IRQn;
	nvicInit.NVIC_IRQChannelPreemptionPriority = 0;
	nvicInit.NVIC_IRQChannelSubPriority = 1;
	nvicInit.NVIC_IRQChannelCmd = ENABLE;
	NVIC_Init(&nvicInit);

	USART_Cmd(USART1, ENABLE);
}

bool uartSendBlock(const uint8_t *data, size_t size) {
	bool rc = fifoWrite(&USART1_transmitBuffer, data, size);

	if (rc && USART1_transmitBuffer.status == STATUS_IDLE) {
		// On a ajouté les octets de 'data' au tampon d'émission.
		// Si l'émission était déjà en cours, ceux-ci seront traités
		// dans la foulée. Par contre, si l'émetteur était en attente,
		// nous devons "amorcer la pompe" en envoyant le premier octet
		// nous-mêmes, la routine d'interruption traitera les suivants.
		USART1_transmitBuffer.status = STATUS_SENDING;
		uint8_t c;
		fifoRead(&USART1_transmitBuffer, &c, 1);
		// Voir commentaire dans USART1_IRQHandler().
		USART_ITConfig(USART1, USART_IT_TXE, ENABLE);
		USART_SendData(USART1, c);
	}

	return rc;
}

bool uartGetBlock(uint8_t *data, size_t size) {
	return fifoRead(&USART1_receiveBuffer, data, size);
}

bool uartIsTransmissionComplete() {
	return USART1_transmitBuffer.status == STATUS_IDLE;
}

bool uartTransmitBufferHasBytesFree(size_t bytes) {
	return fifoBytesFree(&USART1_transmitBuffer) >= bytes;
}

int uartReceiveBufferBytes() {
	return fifoBytesUsed(&USART1_receiveBuffer);
}

void uartFlushReceiveBuffer() {
	fifoClear(&USART1_receiveBuffer);
}
